# -*- coding: utf-8 -*-

import argparse
import rasterio
import numpy as np
import datetime
import os


def parse_arguments():
    parser = argparse.ArgumentParser(prog='pmax.py',description="Create IMGs with MAX. DMP/NPP from LoRes METEO-DATA")

    parser.add_argument("-i", "--in_img_list", type=str, help="file with list of inputfiles",
                        required=True)
    parser.add_argument("-o","--out_file", type=str, help="out file path",
                        required=True)
    parser.add_argument( "--in_nodata", type=int, help="nodata for input files",
                        required=True)
    parser.add_argument("--out_nodata", type=int, help="nodata for output file",
                        required=True)
    parser.add_argument("--out_meta_units", type=str, help="units metatag in the output",
                        required=False)
    parser.add_argument("--out_meta_info", type=str, help="Info metatag in the output",
                        required=False)
    return parser.parse_args()

def validate_inputs(in_img_list):
    bbox_list=[]
    shape_list=[]
    for img in in_img_list:
        with rasterio.open(img) as src:
            shape_list.append(src.shape)
            bbox_list.append(src.bounds)

            #validating blocksize
            blockxsize = int(src.profile["blockxsize"])
            blockysize = int(src.profile["blockysize"])
            # if blockxsize != meteo_settings_dict["input_blockXsize"] or blockysize != meteo_settings_dict["input_blockYsize"]:
            #     raise Exception ("Input blocksize %s,%s does not match the expected ones %s,%s"
            #                      % (blockxsize,
            #                         blockysize,
            #                         meteo_settings_dict["input_blockXsize"],
            #                         meteo_settings_dict["input_blockYsize"]))

    #validate if all items in the list are the same, by using set
    shape_set = set(shape_list)
    bbox_set = set(bbox_list)
    if len(shape_set)!= 1: raise Exception ('Shapes of the in images has differences : %s' % shape_list)
    if len(bbox_set) != 1: raise Exception('Shapes of the in images has differences : %s' % bbox_list)

    total_number_of_blocks = round(int(shape_list[0][0])/blockxsize * int(shape_list[0][1])/blockysize)
    if total_number_of_blocks == 1: print("-Warn- Total number of blocks is 1! For big images this can couse memory" \
                                               " issues. To solve this, provide is tiff with blocks of eg 256:256")
    return total_number_of_blocks

def main():
    args = parse_arguments()
    #create list of inputfiles
    with open(args.in_img_list,'r') as f:
        in_lst = f.readlines()
    in_lst= list(map(lambda in_file: in_file.strip(), in_lst)) #remove line breaks from list
    in_lst.sort()
    # s1s=[]
    # for s1 in s1_in_lst:
    #     s1s.append(join(in_dir, s1['filename']))
    start_composite(in_lst, args.out_file,
                    in_nodata=args.in_nodata,
                    out_nodata=args.out_nodata,
                    out_meta_info=args.out_meta_info,
                    out_meta_units=args.out_meta_units)

def start_composite(in_list, out_file, in_nodata=-1, out_nodata=-32768, out_meta_info=None, out_meta_units=None):
    os.environ['GDAL_CACHEMAX'] = '1500000000'
    #read block per block
    total_number_of_blocks = validate_inputs(in_list)

    #use the first in image to get the tif profile
    with rasterio.open(in_list[0]) as src_sr:
        sr_profile = src_sr.profile
        #set output
        profile = sr_profile
        profile.update(
            dtype=rasterio.int16,
            count=1,
            compress='lzw',
            nodata=out_nodata,
            BIGTIFF='YES'
        )

    #open all files for blockprocessing
    print ("Opening files for %s " % out_file)
    try:
        if os.path.exists(out_file):
            os.remove(out_file)
    except:
        print ("Unable to remove outfile first")
    with rasterio.open(out_file, 'w', sharing=False,  **profile) as dst:
        for ji, window in dst.block_windows(1):
            in_array=[]
            #read the block of all inputs an put it into a multi dimensional array
            for in_file in in_list:
                #create a multi dim numpy array
                with rasterio.open(in_file, sharing=False) as src_sr:
                    data_in = src_sr.read(1, window=window)
                    in_array.append(data_in)
            in_array = np.array(in_array)   #convert regular array into numpy array

            #calculate mean
            in_masked=np.ma.masked_equal(in_array, in_nodata)   #create multi masked array
            mean = in_masked.mean(0)    #create mean over axis=0
            mean_int = np.floor(mean + 0.5).astype(np.int16)  # use rounding half up method

            #convert masked_array to numpy array and fill the masked pixels with nodata
            mean = mean_int.filled(fill_value=out_nodata)
            dst.write(mean, 1, window=window)
        dst.nodata = out_nodata

        #updating the metadata tags in the output geotif
        in_fname_lst=list(map(lambda full_path: os.path.basename(full_path), in_list))
        in_fname_lst.sort()
        dst.update_tags(File_creation=datetime.datetime.now(),
                        creator='VITO',
                        Available_inputs=len(in_list),
                        Used_inputs=', '.join(in_fname_lst)
                        )
        if out_meta_units:
            dst.update_tags(Units=out_meta_units)
        if out_meta_info:
            dst.update_tags(Info=out_meta_info)

    print("Finished successfully with writing %s" % out_file)

if __name__ == "__main__":
    main()
