# -*- coding: utf-8 -*-
from math import exp, floor
import argparse
import rasterio
import numpy as np
import numpy.ma as ma
import os

from frame_s2.config.FRAME_settings import lv2_settings
from frame_s2.postproc_helper import tif_update_metadata

def parse_arguments():
    parser = argparse.ArgumentParser(prog='fapar.py',description="Derive fAPAR from NDVI for FRAME-Project")

    parser.add_argument("--in_file", type=str, help="NDVI (Byte/UNIflagged, smoothed or not",
                        required=True)
    parser.add_argument("--coefficients", nargs='+', type=float, help="Comma-separated list of Polynomial Coefficients ",
                        required=True)
    parser.add_argument("--NDVImin", type=str, help="Physical NDVI-value for fAPAR=0.0, Pixels with NDVI-values below this threshold receive FAPAR=0.0, default=None(automatically calculated)",
                        required=False,
                        default=None)
    parser.add_argument("--out_fapar", type=str, help="FAPAR OUT",
                        required=True)

    return parser.parse_args()

def main():
    args = parse_arguments()
    start_fapar(args.in_file, args.coefficients, args.out_fapar, args.NDVImin)


def calc_fapar(LUT, ndvi_ma):
    fapar_ma = LUT[ndvi_ma]
    return fapar_ma


def validate_input(in_img):
    bbox_list=[]
    shape_list=[]
    with rasterio.open(in_img) as src:
        shape_list.append(src.shape)
        bbox_list.append(src.bounds)

        #validating blocksize
        blockxsize = int(src.profile["blockxsize"])
        blockysize = int(src.profile["blockysize"])

    total_number_of_blocks = round(int(shape_list[0][0])/blockxsize * int(shape_list[0][1])/blockysize)
    if total_number_of_blocks == 1: print("-Warn- Total number of blocks is 1! For big images this can couse memory" \
                                               " issues. To solve this, provide is tiff with blocks of eg 256:256")
    return total_number_of_blocks

def createLUT(NDVImin, coefs):
    #prepare
    if NDVImin == None:
        NDVImin = lv2_settings['NDV_OFFSET'] + 0 * lv2_settings['NDV_SCALE']    #default

    yp = 2.0
    for j in range(250, 0, -1):
        x = lv2_settings['NDV_OFFSET'] + lv2_settings['NDV_SCALE'] * j
        y = coefs[0]
        for i in range(1, len(coefs)):
            y = y + coefs[i] * pow(x, i)
        if (y <= 0.0 or y >= yp):
            NDVImin = x
            break
        yp = y

    #create LUT-table NDVI-FAPAR
    # ---------------------------------------
    # C1.LUT with FAPAR for all NDVI-values (both in terms of digital V
    # ---------------------------------------
    LUT = []
    for c in range(0, 251):
        x = lv2_settings['NDV_OFFSET'] + lv2_settings['NDV_SCALE'] * c  # x = Physical NDVI

        if (x <= NDVImin):
            LUT.append(0)
            continue

        y = coefs[0]
        for i in range(1,len(coefs)):
            y = y + coefs[i] * pow(x, i)
        y = max(0.0, y); y = min(1.0, y); # Keep fAPAR y in range[0.0 - 1.0]
        y = (y - lv2_settings['FPR_OFFSET']) / lv2_settings['FPR_SCALE']
        LUT.append(int(floor(0.5 + y)))

    LUT.extend([251, 252, 253, 254, 255])
    LUT = np.array(LUT, np.ubyte)
    return LUT

def start_fapar(in_NDVI, coefs, out_FAPAR, NDVImin=None):
    #set gdal cache parameter on the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = lv2_settings['GDAL_CACHEMAX']

    LUT = createLUT(NDVImin, coefs)

    #read block per block
    total_number_of_blocks = validate_input(in_NDVI)
    with rasterio.open(in_NDVI) as src_sr:
        sr_profile = src_sr.profile
        #set output
        profile = sr_profile
        profile.update(
            dtype=rasterio.uint8,
            count=1,
            compress='lzw',
            BIGTIFF='YES'
        )

    #open all files for blockprocessing
    with rasterio.open(out_FAPAR, 'w', **profile) as dst:
        with rasterio.open(in_NDVI) as src_sr:
            for ji, window in src_sr.block_windows(1):
                ndvi = src_sr.read(1, window=window)
                ndvi_ma = ma.masked_greater(ndvi, 250)

                fapar = calc_fapar(LUT, ndvi_ma)
                #write result to disk
                dst.write(fapar.astype(rasterio.uint8), 1, window=window)

        dst.nodata = 255

    #update tif metadata
    out_FAPAR = tif_update_metadata(out_FAPAR,
                                    scale=lv2_settings['FPR_SCALE'],
                                    offset=lv2_settings['FPR_OFFSET'],
                                    description=lv2_settings['FPR_DESCRIPTION'],
                                    units=lv2_settings['FPR_UNITS'],
                                    flags=lv2_settings['FPR_FLAGS']
                                    )
    print("Finished successfully with writing %s" % out_FAPAR)
    return out_FAPAR


if __name__ == "__main__":
    main()
