import argparse
from pyspark import SparkContext
from pyspark import SparkConf
import subprocess
import os
from os.path import join
import traceback
import shutil

from frame_s2.config.FRAME_settings import lv2_settings
from frame_s2.postproc_helper import tif_update_metadata

def parse_arguments():

    parser = argparse.ArgumentParser(prog='smooth_all_tiles.py')

    parser.add_argument("--dekad", type=str, help="dekad to process",
                        required=True)
    parser.add_argument("--mask", type=str, help="the mask",
                        required=True)
    parser.add_argument("--input_dir", type=str, help="input folder",
                        required=True)
    parser.add_argument("--output_dir", type=str,
                        help="output folder",
                        required=True)
    parser.add_argument("--tiledir_posix", type=str, help="last foldername where the smoothed data is located, for NRT it is the previous dekad ",
                        required=False)
    parser.add_argument("--tile_list", type=str,
                        help="file with tiles",
                        required=True)
    return parser.parse_args()


def doClip(roi, infile, outdir):
    outfile = join(outdir, os.path.basename(infile))
    gdal_options = "-co 'BLOCKXSIZE=256' -co 'BLOCKYSIZE=256' -co 'TILED=YES' -co 'COMPRESS=LZW'"
    ulx = lv2_settings["ROIS"][roi]["left"]
    uly = lv2_settings["ROIS"][roi]["upper"]
    lrx = lv2_settings["ROIS"][roi]["right"]
    lry = lv2_settings["ROIS"][roi]["lower"]
    cmd = 'gdal_translate -projwin %s %s %s %s %s %s %s' % (ulx, uly, lrx, lry, gdal_options, infile, outfile)
    out, err, errNr = run_command(cmd)
    if errNr:
        print ("Clipping with gdal failed.")
        print(out)
        print(err)
        raise Exception("Unable to run %s" % cmd)
    return outfile


def run_command(cmd):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = p.communicate()
    err_nr = p.returncode
    return str(stdout), str(stderr), err_nr

def mosaic_and_clip(tile_list_file, dekad, output_dir, input_dir, mask, tiledir_posix, roi, clip=True ):
    mosaicdir = join(output_dir, 'tmp')
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    if not os.path.exists(mosaicdir):
        os.mkdir(mosaicdir)

    #remove if already existing
    # #composite_ALB_2111.tif
    for producttype in ['ALB','QUA','NDV']:
        outfile= os.path.join(mosaicdir, 'composite_%s_%s.tif') % (producttype, dekad)
        if os.path.exists(outfile):
            print ("Removing file first: %s" % outfile)
            os.remove(outfile)

        print ("Creating empty mosaic file, copy from mask %s" % outfile )
        try:
            shutil.copy(mask, outfile)
        except:
            raise Exception('Unable to create empty mosaic file %s' % outfile)


    os.environ['GDAL_CACHEMAX'] = lv2_settings['GDAL_CACHEMAX']
    #first step : MOSAICKING
    with open(tile_list_file , 'r') as f:
        tile_list = f.readlines()

    for tile in tile_list:
        tile = tile.strip()
        tile = tile.replace('"', '')   #remove "

        smoothdir = join(input_dir, tile, tiledir_posix)
        clipdir = join(output_dir)
        cmd = 'python3 -m dias_s2chain.mosaic mosaic %s %s "s2t_@VAR@_@DECAD@.tif" %s %s' % (
            dekad,
            smoothdir,
            mosaicdir,
            mask)
        print(cmd)

        stdout, stderr, err_nr = run_command(cmd)
        if err_nr:
            print (traceback.print_exc())
            print (stdout)
            print (stderr)
            print ("Failing mosaic run %s" % cmd)
            raise Exception("mosaic failed for %s" % smoothdir)


    #second step: CLip & update tif metadata
    for type in ['ALB', 'NDV', 'QUA']:
        infile = join(mosaicdir, 'composite_%s_%s.tif') % (type, dekad)
        if clip:
            out_file = doClip(roi, infile, clipdir)
            try:
                os.remove(infile)
            except:
                raise Exception("Unable to remove %s") % infile
        else:
            try:
                out_file = shutil.move(infile, clipdir)
            except:
                raise Exception("Unable to move %s to %s") % (infile, clipdir)


        #set scale & offset to the tif tags
        out_file = tif_update_metadata(out_file, scale=lv2_settings[type + '_SCALE'], offset=lv2_settings[type + '_OFFSET'])

    # removing tempdir
    try:
        os.rmdir(os.path.dirname(infile))
    except:
        pass

    return 0

def create_dekad_list(start_dekad, end_dekad):
    start_dekad = str(start_dekad)
    end_dekad = str(end_dekad)
    dek_list=[]

    dek=start_dekad
    while dek <= end_dekad:
        print(dek)
        dek_list.append(dek)
        yy = dek[0:2]
        kk = str(int(dek[2:4]) + 1).zfill(2) #increase dekad number and print as 2 digit value
        if kk >= '37':
            yy = int(dek[0:2]) + 1
            kk = '01'

        dek="%s%s" % (yy, kk)
    return (dek_list)

def main():
    args = parse_arguments()
    mosaic_and_clip(args.tile_list, args.dekad, args.output_dir, args.input_dir, args.mask, args.tiledir_posix, args.roi, clip=True)




if __name__ == "__main__":
    main()

