# -*- coding: utf-8 -*-
from math import exp, floor
import argparse
import rasterio
import numpy as np
import numpy.ma as ma
import os
from ctypes import  c_double, c_int, CDLL, POINTER

from frame_s2.config.FRAME_settings import meteo_settings_dict
from frame_s2.postproc_helper import rescale, tif_update_metadata


# lib = CDLL("./executables/Pmax_MEP.so")
# plut_c = lib.Pmax_LUT #c_sum is the name of our C function
# pmax_c = lib.Pmax_PIX
#
# plut_c.argtypes = [c_int, c_double, c_double, c_double, c_double, c_double, c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
# plut_c.restype = None
# pmax_c.argtypes = [c_int, c_int, c_int,c_int ,c_int,c_int, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)]
# pmax_c.restype = c_int



# def call_plut_c(Otyp, ARint, ARslo, PhotEff, CO2act, Tc1, Tc2, TcDelt):
#     lutSize=1401
#     LUTeff = (c_double*lutSize)()
#     LUTco2 = (c_double*lutSize)()
#     LUTar = (c_double*lutSize)()
#     plut_c(Otyp, ARint, ARslo, PhotEff, CO2act, Tc1, Tc2, TcDelt, LUTeff, LUTco2, LUTar )
#     return list(LUTeff), list(LUTco2), list(LUTar)

def plut(Otyp, ARint, ARslo, PhotEff, CO2act, Tc1, Tc2, TcDelt):
    #Constants
    KELVIN   = 273.13
    Rgas     = 8.3144       # Universal gas constant, J/K/mol

                            # For general Monteith-equation :
    ClimEff  = 0.48         # kJ(PAR)/kJ(Totalshortwave), climatic efficiency, McCree, 72
    #PhotEff = 2.54         # Old ERROR: photosynthetic efficiency (gDM/MJ(Absorbed PAR) = mgDM/kJ(APAR)) Wolfsy et al., 93
    #PhotEff = 2.45         # That is the correct default value. But now PhotEff is input parameter!
    gCgDM    = 0.45         # gram carbon/gram DM acc (Atjay et al., 83). Needed to convert DMP/GDMP to NPP/GPP.

                                     # For Normalized Temperature Dependency Factor pT (LUTeff)
    C1       = 21.77         # constant
    DHap     = 52750.        # activation energy (J/mol)
    DHdp     = 211000.       # deactivation energy (J/mol)
    Ds       = 704.98        # entropy of CO2 denaturation equilibrium

                           # FOR C02fert (LUTco2)
    O2       = 20.9          # O2 concentration
    CO2ref   = 281.          # CO2 mixing ratio for reference year 1833 (ppmv = µmol/mol)
    A1 = 2.419e+13           # [%CO2]
    E1 = 59400.0             # [J/mol] (Tavg ge 288.13)    CORRECTED (was 54900.0)
    A2 = 1.976e+22           # [%CO2]
    E2 = 109600.0            # [J/mol] (Tavg lt 288.13)
    A0 = 8240.               # [%O2]
    E0 = 13913.5             # [J/mol]
    At = 7.87e-5
    Et =-42896.9             # [J/mol]  - In RSE-publication : -42869.9

    #General
    #"""""""
    LUTeff=[]
    LUTco2=[]
    LUTar=[]

    #short   i, Ni;
    #double  x, y, z, constant, Tk, Km, K0, tau, A;

    #Conversion ENERGY->MATTER
    #Assumed Solar Radiation is in kJ(TOT)/m2/day [0-32767] (32767 can occur exceptionally, but saturated to 32767)
    #""""""""""""""""""""""""""""""""""""""""""""""""""""""
    constant = ClimEff * PhotEff       # DMP/GDMP: 0-32767  mgDM/m²/day = DgDM/ha/day (Dg=Decagram)
    if int(Otyp) == 2 or int(Otyp) == 4 : constant = constant * gCgDM     # NPP/GPP:  0-14745  mgC /m²/day

    #--------------
    #2. CREATE LUTs
    #--------------
    Ni = int(1 + (Tc2 - Tc1)/TcDelt)         #1401 integer temperature steps (incl. 0)
    for i in range(0, Ni):                            #from Tc1 to Tc2 in steps of TcDelt
        # Temperature-variants
        Tk = (i * TcDelt + Tc1) + KELVIN
        x = Rgas * Tk

        #LUTeff = pT = Normalized temperature dependency factor for photosynthetic efficiency
        y  = C1 - DHap / x
        z  = (Ds * Tk - DHdp) / x
        lutval = constant * exp(y) / (1 + exp(z))
        LUTeff.append(lutval)

        #LUTco2 = CO2fert = Normalized CO2-fertilization factor
        if (Tk >= 288.13) :
            Km  = A1 * exp(-E1 / x)
        else:
            Km  = A2 * exp(-E2 / x)

        K0  = A0 * exp(-E0 / x)
        tau = At * exp(-Et / x)
        y = O2/(2.*tau)
        z = Km * (1 + O2/K0)
        LUTco2.append(((CO2act - y)/(CO2ref - y)) * ((z + CO2ref)/(z + CO2act)))

        #LUTar = A = Fraction lost to Autotrophic Respiration (maintenance). NB: For GDMP/GPP: ARint = ARslo = 0.0   => LUTar[i] = 1.0. Correct.
        #A = -3.049 + 0.01145 * Tk;    #Defaults
        A = ARint + ARslo * Tk
        if(A < 0.) : A = 0.        #A cannot be smaller than zero (occurs below -7°C)
        if(A > 1.) : A = 1.        #A may not be greater than 1 (impossible for normal T-range)
        LUTar.append(1. - A)

    return LUTeff, LUTco2, LUTar

# def calc_pmax_c(sr_unscaled, tmin_unscaled, tmax_unscaled, TcShift, flag, fac, LUTeff, LUTco2, LUTar):
#     x=[]
#     for idx, x in np.ndenumerate(sr_unscaled):
#         sr = sr_unscaled[idx]
#         tmin = tmin_unscaled[idx]
#         tmax = tmax_unscaled[idx]
#         array_type = c_double * len(LUTeff)
#
#         pmax = pmax_c(sr, tmin, tmax, 1400, TcShift, flag, fac,
#                       array_type(*LUTeff),
#                       array_type(*LUTco2),
#                       array_type(*LUTar))
#         np.append(x,pmax)
#     return x


def calc_pmax(sr_unscaled, tmin_unscaled, tmax_unscaled, Otyp, TcShift, LUTeff, LUTco2, LUTar):
    fac = 2.0 if (Otyp == 3) else 1.0
    Ni=len(LUTeff)

    #
    # year = CO2year
    # if (CO2year == 0): year = int(date[0:4]) # year YYYY to compute CO2-level (no variation over months)
    # # CO2act = 1.175 * year - 1988; # Veroustraete, 1994. For year 1994, this gives CO2 = 355.61.
    # CO2act = 2.0775 * year - 3785.783

    #LUTeff, LUTco2, LUTar = plut(Otyp, ARint, ARslo, PhotEff, CO2act, Tc1, Tc2, TcDelt)

    #start processing
    T24 = (tmin_unscaled + tmax_unscaled) /2            # Mean TMP over day + night(NB: 1 = Tmin, 2 = Tmax)
    T12 = 0.25 * tmin_unscaled + 0.75 * tmax_unscaled     # Mean TMP over sunshine period

    i24 = TcShift + np.floor(T24 + 0.5).astype(np.int16)    #use rounding half up method
    # Keep track of saturation
    i24 = ma.where(i24 < 0, 0, i24)
    i24 = ma.where(i24 > Ni, Ni, i24)
    #if (i24 < 0): i24 = 0
    #if (i24 > Ni) : i24 = Ni # Entries[0 - 1400] into the three LUTs

    i12 = TcShift + np.floor(T12 + 0.5).astype(np.int16)    #use rounding half up method
    #if (i12 < 0) : i12 = 0
    #if (i12 > Ni): i12 = Ni
    i12 = ma.where(i12 < 0, 0, i12)
    i12 = ma.where(i12 > Ni, Ni, i12)

    #convert to numpy arrays for calculation x later on
    LUTeff=np.array(LUTeff); LUTco2=np.array(LUTco2); LUTar=np.array(LUTar)

    #calculate result and combine the 3 input masks. the SR mask is different from TMP masks
    x = np.ma.array(sr_unscaled * LUTeff[i12] * LUTco2[i12] * LUTar[i24] / fac,
                    mask=(sr_unscaled.mask | tmin_unscaled.mask | tmax_unscaled.mask)) # NB: If GDMP / GPP: LUTar[i] = 1.0!

    # Keep track of saturation
    x = ma.where(x > 32767, 32767, x)
    x = ma.where(x < 0, 0, x)

    x = np.floor(x + 0.5).astype(np.int16)  # use rounding half up method

    return x

def parse_arguments():
    parser = argparse.ArgumentParser(prog='pmax.py',description="Create IMGs with MAX. DMP/NPP from LoRes METEO-DATA")

    parser.add_argument("--in_SR", type=str, help="NAME IN-IMG(s) with SOLAR RADIATION [kJ/m2/d]",
                        required=True)
    parser.add_argument("--in_TMPMAX", type=str, help="NAME IN-IMG(s) with MAX TEMPERATURE [decigrade]",
                        required=True)
    parser.add_argument("--in_TMPMIN", type=str, help="NAME IN-IMG(s) with MIN TEMPERATURE [decigrade]",
                        required=True)
    parser.add_argument("--out_NPM", type=str, help="NAME OUT-IMG(s) with MAXIMUM DMP/NPP",
                        required=True)
    parser.add_argument("--date1", type=str, help="First date in series    [YYYYMMDD]",
                        required=True)
    parser.add_argument("--date2", type=str, help="Last  date in series    [YYYYMMDD]",
                        required=True)
    parser.add_argument("--period", type=str, help="Period: 1=Day, 10=dekad, 30=month",
                        required=True)
    parser.add_argument("--co2year", type=int, help="CO2 : 0=Variable, YYYY=Fix to year",
                        required=True)
    parser.add_argument("--PhotEff", type=float, help="PhotEff [gDM/MJ Absorbed PAR,def=2.45]",
                        default=2.45,
                        required=False)
    parser.add_argument("--ar_slope", type=float, help="AR-Slope     B (default= 0.01145)",
                        default=0.01145,
                        required=False)
    parser.add_argument("--ar_intercept", type=float, help="AR-Intercept A (default=-3.04900)",
                        default=-3.04900,
                        required=False)
    parser.add_argument("--out_type", type=int,
                        help="1=DMP : Dry Matter Productivity [kgDM/ha/d] - Incl. AR;" \
                        "2=NPP : Net Primary Productivity      [gC/m2/d];" \
                        "3=GDMP: Gross Dry Matter Productivity [kgDM/ha/d] - Excl. AR;" \
                        "4=GPP : Gross Primary Productivity    [gC/m2/d]",
                        default=2,
                        required=False)

    return parser.parse_args()

def unscale_meteo(sr_in, tmin_in, tmax_in):
    sr_unscaled = rescale(sr_in,
                          0,
                          1,
                          max_value_in=32767,
                          min_value_in=-9998,
                          max_value_out=32767,
                          min_value_out=0,
                          offset_in=meteo_settings_dict['SR_OFFSET'],
                          scale_in=meteo_settings_dict['SR_SCALE'],
                          out_data_type=np.double)
    sr_masked = ma.masked_equal(sr_unscaled, -9999)
    tmin_unscaled = rescale(tmin_in,
                          0,
                          1,
                          max_value_in=32767,
                          min_value_in=-9998,
                          max_value_out=32767,
                          min_value_out=0,
                          offset_in=meteo_settings_dict['TMIN_OFFSET'],
                          scale_in=meteo_settings_dict['TMIN_SCALE'],
                          out_data_type=np.single)
    tmin_masked = ma.masked_equal(tmin_unscaled, -9999)
    tmax_unscaled = rescale(tmax_in,
                          0,
                          1,
                          max_value_in=32767,
                          min_value_in=-9998,
                          max_value_out=32767,
                          min_value_out=0,
                          offset_in=meteo_settings_dict['TMAX_OFFSET'],
                          scale_in=meteo_settings_dict['TMAX_SCALE'],
                          out_data_type=np.single)
    tmax_masked = ma.masked_equal(tmax_unscaled, -9999)
    return sr_masked, tmin_masked, tmax_masked

def validate_inputs(in_img_list):
    bbox_list=[]
    shape_list=[]
    for img in in_img_list:
        with rasterio.open(img) as src:
            shape_list.append(src.shape)
            bbox_list.append(src.bounds)

            #validating blocksize
            blockxsize = int(src.profile["blockxsize"])
            blockysize = int(src.profile["blockysize"])
            if blockxsize != meteo_settings_dict["input_blockXsize"] or blockysize != meteo_settings_dict["input_blockYsize"]:
                raise Exception ("Input blocksize %s,%s does not match the expected ones %s,%s"
                                 % (blockxsize,
                                    blockysize,
                                    meteo_settings_dict["input_blockXsize"],
                                    meteo_settings_dict["input_blockYsize"]))

    #validate if all items in the list are the same, by using set
    shape_set = set(shape_list)
    bbox_set = set(bbox_list)
    if len(shape_set)!= 1: raise Exception ('Shapes of the in images has differences : %s' % shape_list)
    if len(bbox_set) != 1: raise Exception('Shapes of the in images has differences : %s' % bbox_list)

    total_number_of_blocks = round(int(shape_list[0][0])/blockxsize * int(shape_list[0][1])/blockysize)
    if total_number_of_blocks == 1: print("-Warn- Total number of blocks is 1! For big images this can couse memory" \
                                               " issues. To solve this, provide is tiff with blocks of eg 256:256")
    return total_number_of_blocks

def main():
    args = parse_arguments()
    start_pmax(args.in_SR, args.in_TMPMIN, args.in_TMPMAX, args.out_NPM, args.date1, args.date2, args.out_type, args.ar_intercept, args.ar_slope, args.PhotEff, args.co2year)

#2, 0.5, 0, 1, 0
def start_pmax(in_SR, in_TMPMIN, in_TMPMAX, out_NPM, date1, date2, out_type, ar_intercept, ar_slope, photEff, co2year):
    #set gdal cache parameter on
    # the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = meteo_settings_dict['GDAL_CACHEMAX']

    #general stuf
    Tc1 = -80
    Tc2 = 60
    TcDelt = 0.1 # Temperatures in °C: Range and step widt.NVER CHANGE TcDelt !!!
    TcShift = int(floor(0.5 - Tc1 / TcDelt))

    #CREATE LUT
    year = co2year
    if (co2year == 0): year = int(date1[0:4]) # year YYYY to compute CO2-level (no variation over months)
    # CO2act = 1.175 * year - 1988; # Veroustraete, 1994. For year 1994, this gives CO2 = 355.61.
    CO2act = 2.0775 * year - 3785.783

    LUTeff, LUTco2, LUTar = plut(out_type, ar_intercept, ar_slope, photEff, CO2act, Tc1, Tc2, TcDelt)


    #read block per block
    total_number_of_blocks = validate_inputs([in_SR, in_TMPMAX, in_TMPMIN])
    with rasterio.open(in_SR) as src_sr:
        sr_profile = src_sr.profile
        #set output
        profile = sr_profile
        profile.update(
            dtype=rasterio.int16,
            count=1,
            compress='lzw',
            BIGTIFF='YES')

    # remove output if it exists
    print ("Opening files for file %s" % out_NPM)
    try:
        if os.path.exists(out_NPM):
            os.remove(out_NPM)
    except:
        print ("Unable to remove outfile first")
    # open all files for blockprocessing
    with rasterio.open(out_NPM, 'w', sharing=False, **profile) as dst:
        with rasterio.open(in_SR, 'r', sharing=False) as src_sr:
            with rasterio.open(in_TMPMIN, 'r', sharing=False) as src_tmin:
                with rasterio.open(in_TMPMAX, 'r', sharing=False) as src_tmax:
                    for ji, window in src_sr.block_windows(1):
                        sr = src_sr.read(1, window=window)
                        tmin = src_tmin.read(1, window=window)
                        tmax = src_tmax.read(1, window=window)

                        #unscale the data for calculations : Convert to Physical units: kJ/m²/day or Decigrades Celsius (NEO: V1608)
                        sr_unscaled, tmin_unscaled, tmax_unscaled = unscale_meteo(sr, tmin, tmax)

                        #sr_unscaled as a different mask then tmin and tmax, calculate the overlapping mask of all products and add it to the 3 arrays
                        #overlapping_mask = tmin_unscaled.mask * tmax_unscaled.mask * sr_unscaled.mask

                        npm = calc_pmax(sr_unscaled, tmin_unscaled, tmax_unscaled, out_type, TcShift, LUTeff, LUTco2, LUTar)
                        npm = npm.filled(fill_value=-1)
                        #write result to disk
                        dst.write(npm.astype(rasterio.int16), 1, window=window)


        dst.nodata = -1

    #setting scale and offset to tif
    out_NPM = tif_update_metadata(out_NPM, scale=meteo_settings_dict['NPM_SCALE'], offset=meteo_settings_dict['NPM_OFFSET'])
    print("Finished successfully with writing %s" % out_NPM)


if __name__ == "__main__":
    main()