import argparse
from pyspark import SparkContext
from pyspark import SparkConf
import subprocess
import os
from os.path import join
import datetime
import traceback

from frame_s2.lv1_import_and_mask import start_import_and_mask_lv1
from frame_s2.smooth15 import start_smooth15
from frame_s2.qua_adapt_lv1 import start_qua_adapt
from frame_s2.fapar import start_fapar
from frame_s2.config.FRAME_settings import lv1_settings
from frame_s2.postproc_helper import dateToDekad, getHistoricalDekadList, dekadToStartDate, create_quicklook

#start docker containter
# docker run -v /data/MEP_Projects/Frame:/data/projects/Frame:ro -v ~/Private/Frame/output:/data/output:rw -it frame_smooth
#frame_s2smooth --output-end 2021-03-10 --input-end 2021-03-10 --input-dekads 36 --output-dekads 6 --output-template s2t_{prd}_{dekad} --mask /data/projects/Frame/nrt/S2/REF/S2_SRI_mask.tif /data/projects/Frame//nrt/S2/TILES/44PMQ/TIF/ /data/output/TILES/44PMQ/T10/2107/ --dump-metadata /data/output/logs/44PMQ_2107.log
#

def parse_arguments():

    parser = argparse.ArgumentParser(prog='postprocess_all_lv1.py')

    parser.add_argument("--input_folder_lst", type=str,
                        help="list of modis folders(fullpath, separated by newline) to process.",
                        required=True)
    parser.add_argument("--root_dir", type=str,
                        help="modis root folder, with subdirs=[alb,ndv,fpr,...]",
                        required=True)
    parser.add_argument("--action", type=str,
                        help="process to launch [IMPORT|SMOOTH_NDV|SMOOTH_ALB|QUA_ADAPT]",
                        required=True)
    return parser.parse_args()




def run_command(cmd):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = p.communicate()
    err_nr = p.returncode
    return str(stdout), str(stderr), err_nr


def process_file( folder_name, root_dir, action ):
    os.environ['GDAL_CACHEMAX'] = lv1_settings['GDAL_CACHEMAX']
    ql_width = lv1_settings['ROIS']['AFR']['ql_width']
    ql_height = lv1_settings['ROIS']['AFR']['ql_height']


    #output_dir = /vitodata/Frame/Lv1/
    folder_name = folder_name.strip()      #in format x/y/z/YYYYMMDD/
    folder_name = folder_name.rstrip('/')      #in format x/y/z/YYYYMMDD/

    ##SMOOTHING Parameters
    end_date_yyyymmdd = os.path.basename(folder_name)
    try:
        datetime.datetime.strptime(end_date_yyyymmdd, "%Y%m%d")
    except ValueError as err:
        raise Exception("This folder (%s) does not has the yyyymmdd format" % folder_name)
    current_dekad = dateToDekad(end_date_yyyymmdd)
    start_in_dekad = getHistoricalDekadList(current_dekad, 36)[35]
    start_out_dekad = getHistoricalDekadList(current_dekad, 6)[5]
    start_in_date_yyyymmdd = dekadToStartDate(start_in_dekad)
    start_out_date_yyyymmdd = dekadToStartDate(start_out_dekad)

    # we use input start date for historical processing and input-dekads for nrt processing
    if action.upper() == 'IMPORT':
        print ("Importing and masking = %s" % folder_name)
        start_import_and_mask_lv1(root_dir, lv1_settings['LS_MASK_FILE'], folder_name, lv1_settings['ALB_NORMALISATION'])
    #NDVI + QUA
    elif action.upper() == 'SMOOTH_NDV':
        print ("Smoothing NDVI for %s" % folder_name)
        start_smooth15( 1,   join(root_dir, 'ndv', 'msk', 'L1_NDVI_'), '', start_in_date_yyyymmdd,  end_date_yyyymmdd, 0, '0.004 -0.08 0 250',
                     join(root_dir, 'ndv', 'smo', 'L1_NDVI_'), '', start_out_date_yyyymmdd, end_date_yyyymmdd,
                     join(root_dir, 'ndv', 'quaDay', 'L1_QUAL_'),
                     0, 0.15, 99,  0.025)
        # create quicklook
        for dekad in getHistoricalDekadList(current_dekad, 6):
            create_quicklook(join(root_dir, 'ndv', 'smo', 'L1_NDVI_' + dekad + '.tif'), 'ndvi.txt', height=ql_height, width=ql_width)

    elif action.upper() == 'SMOOTH_ALB':
        #ALBEDO #TODO the 200 max value kan ook 140 zijn
        print ("Smoothing ALB for %s" % folder_name)
        start_smooth15(0,   join(root_dir, 'alb', 'msk', 'L1_ALB_'), '', start_in_date_yyyymmdd,  end_date_yyyymmdd, 0, '0.005 0 0 200',
                     join(root_dir, 'alb', 'smo', 'L1_ALB_'), '',  start_out_date_yyyymmdd, end_date_yyyymmdd, 0,
                     0, 0.00, 99, 0.025)

    elif action.upper() == 'QUA_ADAPT':
        print("Running QUA_ADAPT")
        start_qua_adapt(join(root_dir, 'ndv', 'quaDay', 'L1_QUAL_'),
                        join(root_dir, 'ndv', 'qua', 'L1_QUAL_'),
                        start_out_dekad, current_dekad)

    elif action.upper() == 'FAPAR':
        #FAPAR generation
        # fapar
        in_dir = join(root_dir, 'ndv', 'smo')  # previous nrt versions(until currentdekad-5) will be overwritten
        for dekad in getHistoricalDekadList(current_dekad, 6):
            in_file = join(in_dir, 'L1_NDVI_%s.tif' % dekad)
            fapar_file = join(root_dir, 'fpr', 'smo', 'L1_FPR_%s.tif' % dekad)
            start_fapar(in_file, lv1_settings['FPR_POLYNOMIAL_COEFFICIENTS_MODIS'], fapar_file)
    else:
        raise Exception ("Action is not valid. is %s, should be 'IMPORT', 'FAPAR', 'SMOOTH_NDV', 'SMOOTH_ALB', 'QUA_ADAPT'" % action)

def main():
    args = parse_arguments()
    print(args)

    sc_conf = SparkConf()
    sc = SparkContext(conf=sc_conf)


    if not os.path.exists(args.input_folder_lst):
        raise Exception("The input folder list does not exist : %s" % args.input_folder_lst)
    try:
        with open(args.input_folder_lst) as f:
            in_files = f.readlines()
        #in_files = list(map(lambda fname: fname.replace('"',''), in_files))  # remove " around tiles
        number_of_files = len(in_files)
        if number_of_files > 0:
            #logdir = os.path.join(args.output_dir, 'logs')
            tiles_rdd = sc.parallelize(in_files, len(in_files))
            tiles_rdd.foreach(lambda fname: process_file(fname, args.root_dir, args.action))
        else:
            raise Exception("input list is empty %s" % args.input_folder_lst )

    finally:
        sc.stop()


if __name__ == "__main__":
    main()

