# -*- coding: utf-8 -*-
from math import exp, floor
import argparse
import rasterio
import numpy as np
import numpy.ma as ma
import os
import re
import glob

from frame_s2.postproc_helper import validate_inputs, tif_update_metadata, _str2bool, getDekadLst
from frame_s2.config.FRAME_settings import meteo_settings_dict,lv2_settings

def parse_arguments():
    parser = argparse.ArgumentParser(prog='qua_adapt.py',description="qua_adapt")

    parser.add_argument("--in_qua_day", type=str, help="single QUA_DAY input image, units=days",
                        required=True,
                        default=None)
    parser.add_argument("--out_qua_dek", type=str, help="single QUA_DEK output image, units=dekads",
                        required=True,
                        default=None)
    # parser.add_argument("--in_dekad", type=int, help="dekad in format YYTT",
    #                     required=True)
    parser.add_argument("--previous_out_qua_dek", type=str, help="Use Previous QUADekad file as input, if None(default), it won't be used",
                        default=None)
    parser.add_argument("--S2", type=_str2bool, help="S2|(Modis|PROBA (smooth15)", default=True,
                        required=False)


    return parser.parse_args()


def main():
    args = parse_arguments()
    start_qua_adapt(args.in_qua_day, args.out_qua_dek, args.previous_out_qua_dek, args.S2)

def start_qua_adapt( quaday_img, current_dek_img, previous_out_qua_dek, S2):
    #set gdal cache parameter on the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = meteo_settings_dict['GDAL_CACHEMAX']

    #initialize
    hist_gap = 0        #gap beteween current dekad and previous_out_qua_Dek
    infiles_2_validate = []



    #get inital image
    infiles_2_validate=[quaday_img]
    if previous_out_qua_dek != None:
        try:
            hist_dekad = re.search('[0-9]{4}', os.path.basename(previous_out_qua_dek)).group(0)
            current_dekad = re.search('[0-9]{4}', os.path.basename(current_dek_img)).group(0)
        except:
            raise Exception('Unable to get dekad out of filenames [%s,%s]' %(os.path.basename(previous_out_qua_dek),os.path.basename(current_dek_img)))
        #hist_gap = int(current_dekad) - int(hist_dekad)
        hist_gap = len(getDekadLst(hist_dekad, current_dekad)) - 1  #number of dekads between hist and current dekad
        #use previous dekad to get initial values.
        src_prev = rasterio.open(previous_out_qua_dek)
        infiles_2_validate.append(previous_out_qua_dek)

    #validate all inputs
    validate_inputs(infiles_2_validate)

    #creating profile
    with rasterio.open(quaday_img) as src:
        src_profile = src.profile
        # set output
        profile = src_profile
        profile.update(
            dtype=rasterio.uint8,
            count=1,
            compress='lzw',
            BIGTIFF='YES')



    # start processing single file
    with rasterio.open(quaday_img) as src_curr:
        with rasterio.open(current_dek_img, 'w', **profile) as dst:
            for ji, window in src_curr.block_windows(1):
                current_day = src_curr.read(1, window=window)

                if previous_out_qua_dek != None:
                    prev = src_prev.read(1, window=window)
                else:
                    prev = np.full(current_day.shape, 251, np.uint8)        #use 251, then prev will be skipped and only the day image is used, if 254, it will be masked


                current_dek = calc_qua_using_hist(current_day, prev, hist_gap, S2=S2)
                #calculate without previous image, good enough for S2, not for PROBA and modis(Lv1)
                #current_dek = calc_qua(current_day)
                dst.write(current_dek, window=window, indexes=1)

                #print('writing %s with %s' % (current_dek_img, window))
                #      previous_out_qua_dek = current_dek_img
            dst.nodata = 255
            dst.close()
        src_curr.close()

    if previous_out_qua_dek != None:
        src_prev.close()

    #update tif metadata
    current_dek_img = tif_update_metadata(current_dek_img,
                                          scale=lv2_settings['QUA_SCALE'],
                                          offset=lv2_settings['QUA_OFFSET'],
                                          description=lv2_settings['QUA_DESCRIPTION'],
                                          units=lv2_settings['QUA_UNITS'],
                                          flags=lv2_settings['QUA_FLAGS']
                                          )
    print("Finished successfully with writing %s" % current_dek_img)
    return current_dek_img

def calc_qua_using_hist(current_day, hist, hist_gap, keep250=True, S2=True):
    #all values > 251 will be flagged to 254
    # values = 251 should remain
    msk_255_day = np.where(current_day == 255, 1, 0)
    #convert acceptable observations to good observations
    current_day_masked = ma.masked_greater(current_day, 252)
    if keep250:
        current_day_masked = ma.where(current_day_masked == 250, 0, current_day_masked) #special case: if there is a small number of observations in the smoother, it will set this to 250. We will hard code this to 0

    #mask missing land pixels
    #current_day_masked_values = ma.masked_less(current_day, 251)                #expressed in days
    prev_dek_masked = ma.masked_greater(hist, 252)                              #expressed in dekads
    current_dek_using_hist = ma.where(prev_dek_masked < 251, prev_dek_masked + hist_gap, prev_dek_masked)                                   #QUA_DEK = QUA_DEK_HIST + GAP ( do not increase missing land data)
    current_dek_using_hist = ma.where(prev_dek_masked < 251, np.minimum(current_dek_using_hist, 240), current_dek_using_hist)                ## solve saturation, but keep 251
    prev_dek_masked = None

    dif = 0 if S2 else 1  #qua_day for smooth15(PROBA) is never 10(1dekad). At least these are 20/21 days, so we do minus 1 when using smooth15 qua results
    current_dek_using_current_day = ma.where(current_day_masked <= 240, np.maximum((current_day_masked / 10 - dif), 0).astype(np.ubyte) ,  current_day_masked)        # if not S2, do minus 1, but set to 0 if results become negative, keep the day masked values as they were


    #from here, all calculations are expressed in dekads
    #FOR S2 data. All qua_values>150 days are flagged as 251!!!!! Also NDVI is flagged!
    #current_dek_masked = ma.minimum(current_dek_using_current_day, current_dek_using_hist) #
    #current_dek_masked = ma.where(current_dek_using_current_day < hist_gap, current_dek_using_current_day, prev_dek_masked + hist_gap)   #use prev value if current image has no matches
    current_dek_masked = ma.where(current_dek_using_current_day == 251, current_dek_using_hist , current_dek_using_current_day)   # solve saturation, but keep 251
    current_dek_masked = ma.where(current_day_masked == 240, current_dek_using_hist,
                                  current_dek_masked)  #if day image is saturated, use hist_dekad
    #current_dek_masked = ma.where(current_dek_using_current_day == 251, 251, current_dek_masked)          #keep current_day value of 251
    current_dek = current_dek_masked.filled(fill_value=254)    # reset flagged values
    current_dek = np.where(msk_255_day == 1, 255, current_dek)     #set 255 pixels again
    current_day_masked = None
    msk_255_day = None
    return current_dek

def calc_qua_herman_like(current_day, prev):
    current_day_masked = ma.masked_greater(current_day, 250)
    current_day_masked = ma.where(current_day_masked == 250, 0, current_day_masked) #special case: if there is a small number of observations in the smoother, it will set this to 250. We will hard code this to 0
    current_day_masked_values = ma.masked_less(current_day, 251)
    prev_masked = ma.masked_greater(prev, 251)
    prev_masked = ma.where(prev_masked == 250, 0, prev_masked)

    current_day_masked = (current_day_masked / 10).astype(np.ubyte)  # from days to dekads

    #use previous result as
    current_day_masked = ma.where(current_day_masked == 0, 0, prev_masked + 1)
    current_day_masked = ma.where(current_day_masked >= 240, 240,
                                  current_day_masked)  # solve saturation
    current_day = current_day_masked.filled(fill_value=0) + current_day_masked_values.filled(
        fill_value=0)  # rebuilding the array with nodata values filled in
    # current_day = current_day_masked.filled(fill_value=flag_out)
    return  current_day

def calc_qua(current_day):
    current_day_masked = ma.masked_greater(current_day, 250)
    current_day_masked = ma.where(current_day_masked == 250, 0, current_day_masked )   #special case: if there is a small number of observations in the smoother, it will set this to 250. We will hard code this to 0
    current_day_masked_values = ma.masked_less(current_day, 251)
    current_day_masked = (current_day_masked / 10).astype(np.ubyte)  # from days to dekads

    current_day_masked = ma.where(current_day_masked >= 240, 240,
                                  current_day_masked)  # solve saturation
    current_day = current_day_masked.filled(fill_value=0) + current_day_masked_values.filled(
                    fill_value=0)  # rebuilding the array with nodata values filled in
                                    # current_day = current_day_masked.filled(fill_value=flag_out)
    return  current_day
if __name__ == "__main__":
    main()